{$B-}

unit mn_kazal;

interface

uses mn_abstr; {mnozica, realizirana z linearnim seznamom}

type
  kElKazalci=^elKazalci;
  elKazalci=record {element linearnega seznama}
    vsebina: kVsebina; {kazalec na element mnozice}
    naslednji: kElKazalci; {kazalec na naslednji element linearnega seznama}
  end;
  kMnKazalci=^mnKazalci;
  mnKazalci=object (mnozica) {glej mn_abstr}
    prvi: kElKazalci; {kazalec na prvi element linearnega seznama (ki je prazen - glava)}
    destructor mnDest; virtual;
    procedure naredi; virtual;
    procedure izvoziU (mn: kMnozica; preveri, iter: boolean); virtual;
    procedure izvoziP (mn, drugaMn: kMnozica; iter: boolean); virtual;
    function dodaj (vs: kVsebina; preveri, iter: boolean): boolean; virtual;
    function brisi (vs: kVsebina): boolean; virtual;
    function pripada (vs: kVsebina; iter: boolean): boolean; virtual;
    function unija (mn1, mn2: kMnozica; iter: boolean): boolean; virtual;
    function presek (mn1, mn2: kMnozica; iter: boolean): boolean; virtual;
  end;

implementation

destructor mnKazalci.mnDest;
  var trenutni, zac: kElKazalci;
begin
  trenutni:=prvi;
  while trenutni<>nil do begin {na zacetku krajsa seznam}
    zac:=trenutni;
    trenutni:=trenutni^.naslednji;
    if zac<>prvi then dispose (zac^.vsebina, vsDest); {sprosti spomin, ki ga je zasedal element mnozice}
    dispose (zac); {sprosti spomin, ki ga je zasedal element linearnega seznama}
  end;
end;

procedure mnKazalci.naredi;
begin
  tip:=kazalci;
  prvi:=new (kElKazalci); {naredi prvi element, ki je vedno prazen (glava)}
  prvi^.naslednji:=nil;
end;

procedure mnKazalci.izvoziU (mn: kMnozica; preveri, iter: boolean);
  var
    trenutni: kElKazalci;
    zac: kVsebina;
  procedure rekurzija (polozaj: kElKazalci); {procedura za rekurzijo - vsebina podobna kot pri iteraciji}
  begin
    if polozaj<>nil then begin
      zac:=polozaj^.vsebina^.prepisi;
      if not mn^.dodaj (zac, preveri, false) then dispose (zac, vsDest);
      rekurzija (polozaj^.naslednji);
    end;
  end;
begin
  if iter then begin {iterativno ...}
    trenutni:=prvi^.naslednji;
    while trenutni<>nil do begin
      zac:=trenutni^.vsebina^.prepisi; {v zac se zapise kopija trenutnega elementa mnozice}
      if not mn^.dodaj (zac, preveri, true) then dispose (zac, vsDest); {ce dodajanje ne uspe, se zbrise}
      trenutni:=trenutni^.naslednji;
    end;
  end else rekurzija (prvi^.naslednji); {... ali rekurzivno}
end;

procedure mnKazalci.izvoziP (mn, drugaMn: kMnozica; iter: boolean);
  var trenutni: kElKazalci;
  procedure rekurzija (polozaj: kElKazalci); {procedura za rekurzijo - vsebina podobna kot pri iteraciji}
  begin
    if polozaj<>nil then begin
      if drugaMn^.pripada (polozaj^.vsebina, false) then
        mn^.dodaj (polozaj^.vsebina^.prepisi, false, false);
      rekurzija (polozaj^.naslednji);
    end;
  end;
begin
  if iter then begin {iterativno ...}
    trenutni:=prvi^.naslednji;
    while trenutni<>nil do begin
      if drugaMn^.pripada (trenutni^.vsebina, true) then     {ce trenutni element pripada drugaMn, ->}
        mn^.dodaj (trenutni^.vsebina^.prepisi, false, true); {se njegova kopija zapise v mn}
      trenutni:=trenutni^.naslednji;
    end;
  end else rekurzija (prvi^.naslednji); {... ali rekurzivno}
end;

function mnKazalci.dodaj (vs: kVsebina; preveri, iter: boolean): boolean;
  var zac: kElKazalci;
begin
  if (not preveri) or (not pripada (vs, iter)) then begin {ce element ni ze vsebovan ali pa tega ni treba preveriti, se ->}
    zac:=new (kElKazalci);                                {doda za glavo}
    zac^.vsebina:=vs;
    zac^.naslednji:=prvi^.naslednji;
    prvi^.naslednji:=zac;
    dodaj:=true;
  end else dodaj:=false;
end;

function mnKazalci.brisi (vs: kVsebina): boolean;
  var trenutni, zac: kElKazalci;
begin
  brisi:=false;
  trenutni:=prvi;
  while trenutni^.naslednji<>nil do begin
    if trenutni^.naslednji^.vsebina^.jeEnak (vs) then begin
      zac:=trenutni^.naslednji;
      trenutni^.naslednji:=zac^.naslednji; {linearni seznam preveze mimo zbrisanega elementa seznama}
      dispose (zac^.vsebina, vsDest); {sprosti spomin, ki ga je zasedal element mnozice}
      dispose (zac); {sprosti spomin, ki ga je zasedal element linearnega seznama}
      brisi:=true;
      break;
    end;
    trenutni:=trenutni^.naslednji;
  end;
end;

function mnKazalci.pripada (vs: kVsebina; iter: boolean): boolean;
  var trenutni: kElKazalci;
  procedure rekurzija (polozaj: kElKazalci); {procedura za rekurzijo - vsebina podobna kot pri iteraciji}
  begin
    if polozaj<>nil then
      if vs^.jeEnak (polozaj^.vsebina) then
        pripada:=true else rekurzija (polozaj^.naslednji);
  end;
begin
  pripada:=false;
  if iter then begin {iterativno ...}
    trenutni:=prvi^.naslednji;
    while trenutni<>nil do begin
      if vs^.jeEnak (trenutni^.vsebina) then begin {ko prispe do iskanega elementa, zakljuci in vrne true}
        pripada:=true;
        break;
      end;
      trenutni:=trenutni^.naslednji;
    end;
  end else rekurzija (prvi^.naslednji); {... ali rekurzivno}
end;

function mnKazalci.unija (mn1, mn2: kMnozica; iter: boolean): boolean; {enako kot v mn_polje}
  var preveri: boolean;
begin
  if (mn1<>nil) and (mn2<>nil) then begin
    if (mn1<>@self) and (mn2<>@self) then begin
      mnDest;
      naredi;
      preveri:=false;
    end else preveri:=true;
    if (mn1<>@self) then mn1^.izvoziU (@self, preveri, iter);
    if (mn2<>@self) then mn2^.izvoziU (@self, true, iter);
    unija:=true;
  end else unija:=false;
end;

function mnKazalci.presek (mn1, mn2: kMnozica; iter: boolean): boolean;
  var
    zac, trenutni: kElKazalci;
    mn: kMnozica;
  procedure rekurzija (polozaj: kElKazalci); {procedura za rekurzijo - vsebina podobna kot pri iteraciji}
  begin
    if polozaj^.naslednji<>nil then begin
      if not mn^.pripada (polozaj^.naslednji^.vsebina, false) then begin
        zac:=polozaj^.naslednji;
        polozaj^.naslednji:=zac^.naslednji;
        dispose (zac^.vsebina, vsDest);
        dispose (zac);
        rekurzija (polozaj);
      end else rekurzija (polozaj^.naslednji);
    end;
  end;
begin
  if (mn1<>nil) and (mn2<>nil) then begin
    if (mn1=@self) xor (mn2=@self) then begin {ce je ena izmed mnozic preseka mnozica, v katero se presek zapisuje}
      if (mn1=@self) then mn:=mn2 else mn:=mn1; {mn postane mnozica, v katero se ne zapisuje presek}
      if iter then begin
        trenutni:=prvi;
        while trenutni^.naslednji<>nil do begin {kar ne spada v presek, se zbrise iterativno (na enak nacin kot v brisi) ...}
          if not mn^.pripada (trenutni^.naslednji^.vsebina, true) then begin
            zac:=trenutni^.naslednji;
            trenutni^.naslednji:=zac^.naslednji;
            dispose (zac^.vsebina, vsDest);
            dispose (zac);
          end else trenutni:=trenutni^.naslednji;
        end;
      end else rekurzija (prvi); {... ali rekurzivno}
    end;
    if (mn1<>@self) and (mn2<>@self) then begin {ce ni nobena izmed mnozic preseka mnozica, v katero se presek zapisuje, se ->}
      mnDest;                                   {izprazni}
      naredi;
      mn1^.izvoziP (@self, mn2, iter); {dodajo se elementi mn1, ki ne pripadajo mn2 - uporabi se metoda mn1}
    end;
    presek:=true;
  end else presek:=false;
end;

end.