{$B-}

unit mn_kurz; {mnozica, realizirana s kurzorji}

interface

uses mn_abstr;

const
  maxKurzorji=22;

type
  elKurzorji=record {element polja}
    vsebina: kVsebina; {kazalec na element mnozice}
    naslednji: integer; {indeks naslednjega elementa polja}
  end;
  kMnKurzorji=^mnKurzorji;
  mnKurzorji=object (mnozica) {glej mn_abstr}
    polje: array [0..maxKurzorji] of elKurzorji; {polje, v katerem so elementi}
    zadnji: integer; {indeks zadnjega polnega elementa}
    destructor mnDest; virtual;
    procedure naredi; virtual;
    procedure izvoziU (mn: kMnozica; preveri, iter: boolean); virtual;
    procedure izvoziP (mn, drugaMn: kMnozica; iter: boolean); virtual;
    function dodaj (vs: kVsebina; preveri, iter: boolean): boolean; virtual;
    function brisi (vs: kVsebina): boolean; virtual;
    function pripada (vs: kVsebina; iter: boolean): boolean; virtual;
    function unija (mn1, mn2: kMnozica; iter: boolean): boolean; virtual;
    function presek (mn1, mn2: kMnozica; iter: boolean): boolean; virtual;
  end;

implementation

destructor mnKurzorji.mnDest;
  var trenutni: integer;
begin
  trenutni:=polje [0].naslednji;
  while trenutni<>polje [zadnji].naslednji do begin
    dispose (polje [trenutni].vsebina, vsDest); {sprosti spomin, ki so ga zasedali elementi mnozice}
    trenutni:=polje [trenutni].naslednji;
  end;
end;

procedure mnKurzorji.naredi;
  var n: integer;
begin
  tip:=kurzorji;
  zadnji:=0; {doloci, da je konec seznama na 0-tem elementu (ki je vedno prazen - glava)}
  for n:=0 to maxKurzorji do polje [n].naslednji:=(n+1) mod (maxKurzorji+1); {kurzorji kazejo na naslednje elemente polja,
                                                                             zadnji pa na zacetek}
end;

procedure mnKurzorji.izvoziU (mn: kMnozica; preveri, iter: boolean);
  var
    trenutni: integer;
    zac: kVsebina;
  procedure rekurzija (polozaj: integer); {procedura za rekurzijo - vsebina podobna kot pri iteraciji}
  begin
    if polozaj<>polje [zadnji].naslednji then begin
      zac:=polje [polozaj].vsebina^.prepisi;
      if not mn^.dodaj (zac, preveri, false) then dispose (zac, vsDest);
      rekurzija (polje [polozaj].naslednji);
    end;
  end;
begin
  if iter then begin {iterativno ...}
    trenutni:=polje [0].naslednji;
    while trenutni<>polje [zadnji].naslednji do begin
      zac:=polje [trenutni].vsebina^.prepisi; {v zac se zapise kopija trenutnega elementa mnozice}
      if not mn^.dodaj (zac, preveri, true) then dispose (zac, vsDest); {ce dodajanje ne uspe, se zbrise}
      trenutni:=polje [trenutni].naslednji;
    end;
  end else rekurzija (polje [0].naslednji); {... ali rekurzivno}
end;

procedure mnKurzorji.izvoziP (mn, drugaMn: kMnozica; iter: boolean);
  var trenutni: integer;
  procedure rekurzija (polozaj: integer); {procedura za rekurzijo - vsebina podobna kot pri iteraciji}
  begin
    if polozaj<>polje [zadnji].naslednji then begin
      if drugaMn^.pripada (polje [polozaj].vsebina, false) then
        mn^.dodaj (polje [polozaj].vsebina^.prepisi, false, false);
      rekurzija (polje [polozaj].naslednji);
    end;
  end;
begin
  if iter then begin {iterativno ...}
    trenutni:=polje [0].naslednji;
    while trenutni<>polje [zadnji].naslednji do begin
      if drugaMn^.pripada (polje [trenutni].vsebina, true) then     {ce trenutni element pripada drugaMn, ->}
        mn^.dodaj (polje [trenutni].vsebina^.prepisi, false, true); {se njegova kopija zapise v mn}
      trenutni:=polje [trenutni].naslednji;
    end;
  end else rekurzija (polje [0].naslednji); {... ali rekurzivno}
end;

function mnKurzorji.dodaj (vs: kVsebina; preveri, iter: boolean): boolean;
begin
  if ((not preveri) or (not pripada (vs, iter))) and (polje [zadnji].naslednji>0) then begin {ce je prostor in element ni ->}
    zadnji:=polje [zadnji].naslednji;                                                        {ze vsebovan ali pa tega ni ->}
    polje [zadnji].vsebina:=vs;                                                              {treba preveriti, se doda na ->}
    dodaj:=true;                                                                             {konec}
  end else dodaj:=false;
end;

function mnKurzorji.brisi (vs: kVsebina): boolean;
  var
    trenutni, zac: integer;
begin
  brisi:=false;
  trenutni:=0;
  while polje [trenutni].naslednji<>polje [zadnji].naslednji do begin
    if vs^.jeEnak (polje [polje [trenutni].naslednji].vsebina) then begin
      zac:=polje [trenutni].naslednji;
      dispose (polje [zac].vsebina, vsDest); {sprosti spomin, ki ga je zasedal element mnozice}
      if zac=zadnji then zadnji:=trenutni else begin {ce je bil zadnji, zadnji nastavi na prejsnjega}
        polje [trenutni].naslednji:=polje [zac].naslednji; {sicer polne elemente polja preveze mimo zbrisanega}
        polje [zac].naslednji:=polje [zadnji].naslednji; {zbrisanega pa ->}
        polje [zadnji].naslednji:=zac;                   {preveze za zadnjega}
      end;
      brisi:=true;
      break;
    end;
    trenutni:=polje [trenutni].naslednji;
  end;
end;

function mnKurzorji.pripada (vs: kVsebina; iter: boolean): boolean;
  var trenutni: integer;
  procedure rekurzija (polozaj: integer); {procedura za rekurzijo - vsebina podobna kot pri iteraciji}
  begin
    if polozaj<>polje [zadnji].naslednji then
      if vs^.jeEnak (polje [polozaj].vsebina) then
        pripada:=true else rekurzija (polje [polozaj].naslednji);
  end;
begin
  pripada:=false;
  if iter then begin {iterativno ...}
    trenutni:=polje [0].naslednji;
    while trenutni<>polje [zadnji].naslednji do begin
      if vs^.jeEnak (polje [trenutni].vsebina) then begin {ko prispe do iskanega elementa, zakljuci in vrne true}
        pripada:=true;
        break;
      end;
      trenutni:=polje [trenutni].naslednji;
    end;
  end else rekurzija (polje [0].naslednji); {... ali rekurzivno}
end;

function mnKurzorji.unija (mn1, mn2: kMnozica; iter: boolean): boolean; {enako kot v mn_polje}
  var preveri: boolean;
begin
  if (mn1<>nil) and (mn2<>nil) then begin
    if (mn1<>@self) and (mn2<>@self) then begin
      mnDest;
      naredi;
      preveri:=false;
    end else preveri:=true;
    if (mn1<>@self) then mn1^.izvoziU (@self, preveri, iter);
    if (mn2<>@self) then mn2^.izvoziU (@self, true, iter);
    unija:=true;
  end else unija:=false;
end;

function mnKurzorji.presek (mn1, mn2: kMnozica; iter: boolean): boolean;
  var
    zac, trenutni: integer;
    mn: kMnozica;
  procedure rekurzija (polozaj: integer); {procedura za rekurzijo - vsebina podobna kot pri iteraciji}
  begin
    if polje [polozaj].naslednji<>polje [zadnji].naslednji then begin
      if not mn^.pripada (polje [polje [polozaj].naslednji].vsebina, false) then begin
        zac:=polje [polozaj].naslednji;
        dispose (polje [zac].vsebina, vsDest);
        if zac=zadnji then zadnji:=polozaj else begin
          polje [polozaj].naslednji:=polje [zac].naslednji;
          polje [zac].naslednji:=polje [zadnji].naslednji;
          polje [zadnji].naslednji:=zac;
        end;
        rekurzija (polozaj);
      end else rekurzija (polje [polozaj].naslednji);
    end;
  end;
begin
  if (mn1<>nil) and (mn2<>nil) then begin
    if (mn1=@self) xor (mn2=@self) then begin {ce je ena izmed mnozic preseka mnozica, v katero se presek zapisuje}
      if (mn1=@self) then mn:=mn2 else mn:=mn1; {mn postane mnozica, v katero se ne zapisuje presek}
      if iter then begin
        trenutni:=0;
        while polje [trenutni].naslednji<>polje [zadnji].naslednji do begin                {kar ne spada v presek, se ->}
          if not mn^.pripada (polje [polje [trenutni].naslednji].vsebina, true) then begin {zbrise iterativno (na enak ->}
            zac:=polje [trenutni].naslednji;                                               {nacin kot v brisi) ...}
            dispose (polje [zac].vsebina, vsDest);
            if zac=zadnji then zadnji:=trenutni else begin
              polje [trenutni].naslednji:=polje [zac].naslednji;
              polje [zac].naslednji:=polje [zadnji].naslednji;
              polje [zadnji].naslednji:=zac;
            end;
          end else trenutni:=polje [trenutni].naslednji;
        end;
      end else rekurzija (0); {... ali rekurzivno}
    end;
    if (mn1<>@self) and (mn2<>@self) then begin {ce ni nobena izmed mnozic preseka mnozica, v katero se presek zapisuje, ->}
      mnDest;                                   {se izprazni}
      naredi;
      mn1^.izvoziP (@self, mn2, iter); {dodajo se elementi mn1, ki ne pripadajo mn2 - uporabi se metoda mn1}
    end;
    presek:=true;
  end else presek:=false;
end;

end.